"""Module to handle bootstrap of project."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


from jinja2 import Environment, PackageLoader
from rich import print as rprint

from pspm.utils.commands import get_git_user


def _create_package_structure(base_path: Path, safe_name: str) -> None:
    package_init = base_path / "src" / safe_name / "__init__.py"
    package_init.parent.mkdir(parents=True, exist_ok=True)
    package_init.write_text("")


def bootstrap_project(
    path: Path, name: str | None, description: str | None, *, installable: bool
) -> None:
    """Create initial project structure.

    Args:
        path: Path to project
        name: Project name
        description: Project description
        installable: Whether the project is installable
    """
    name = name or path.absolute().name
    description = description or "Describe your project here"

    path.mkdir(exist_ok=True)
    env = Environment(loader=PackageLoader("pspm"), autoescape=True)
    template_names = env.list_templates()
    author = get_git_user()
    safe_package_name = name.lower().replace("-", "_")
    for template_name in template_names:
        template = env.get_template(template_name)
        result = template.render(
            name=name,
            safe_name=safe_package_name,
            description=description,
            installable=installable,
            author=author,
        )
        file_name = template_name.replace(".j2", "")
        file_path = path / file_name
        file_path.write_text(result, encoding="utf-8")
    _create_package_structure(path, safe_package_name)
    rprint(
        f"Initialized project [blue]{name}[/blue] "
        + (f"in [blue]{path.name}[/blue]" if path.name else "")
    )



"""Module to handle dependencies."""

from __future__ import annotations

from pathlib import Path
from typing import Literal

from pspm.entities.installer import BaseInstaller, UVInstaller
from pspm.entities.package_manager import PackageManager
from pspm.entities.pyproject import Pyproject
from pspm.entities.resolver import BaseResolver, UVResolver
from pspm.entities.toml import Toml
from pspm.entities.virtual_env import VirtualEnv


def _get_pyproject_path() -> str:
    path = Path(Path.cwd()) / "pyproject.toml"
    return str(path)


def _get_pyproject() -> Pyproject:
    path = _get_pyproject_path()
    parser = Toml(str(path))
    return Pyproject(parser)


def _get_resolver() -> BaseResolver:
    path = _get_pyproject_path()
    return UVResolver(path)


def _get_installer() -> BaseInstaller:
    return UVInstaller()


def _get_package_manager() -> PackageManager:
    virtual_env = VirtualEnv()
    return PackageManager(
        _get_pyproject(), _get_installer(), _get_resolver(), virtual_env
    )


def install_dependencies() -> None:
    """Install all dependencies and the package itself."""
    package_manager = _get_package_manager()
    package_manager.install()


def manage_dependency(
    action: Literal["add", "remove"], package: str, group: str | None = None
) -> None:
    """Add dependency to pyproject.

    Args:
        action: Action to take can be either add or remove
        package: Package to install
        group: Group to insert package
    """
    package_manager = _get_package_manager()
    package_manager.manage_dependency(action, package, group)


def lock_dependencies(*, update: bool = False) -> None:
    """Lock dependencies."""
    package_manager = _get_package_manager()
    package_manager.compile_requirements(upgrade=update)
    package_manager.install()


def get_version() -> str:
    """Retrive pyproject version.

    Returns:
        Project version.
    """
    pyproject = _get_pyproject()
    return pyproject.version


def change_version(
    new_version: str | None = None,
    bump_rule: Literal["major", "minor", "patch"] | None = None,
) -> str:
    """Change project version.

    Args:
        new_version: Version to change to
        bump_rule: Rule to change version

    Raises:
        ValueError: If neither argument were provided

    Returns:
        Updated version
    """
    pyproject = _get_pyproject()
    if new_version:
        return pyproject.change_version(new_version)
    if bump_rule:
        return pyproject.bump_version(bump_rule)
    error_message = "Missing action to change version"
    raise ValueError(error_message)



from pspm.entities.virtual_env import BaseVirtualEnv
import pytest

from pspm.entities.package_manager import PackageManager
from pspm.entities.pyproject import BasePyproject
from pspm.entities.toml import BaseToml
from pspm.entities.installer import BaseInstaller
from pspm.entities.resolver import BaseResolver
from typing import Any, Literal


class DummyToml(BaseToml):
    def __init__(self, data: dict[str, Any]) -> None:
        self.data = data

    def load(self) -> dict[str, Any]:
        return self.data

    def dump(self, data: dict[str, Any]) -> None:
        self.data = data


class DummyPyproject(BasePyproject):
    def __init__(self, toml_parser: BaseToml) -> None:
        self._parser = toml_parser
        self.added_dependencies: list[str] = []
        self.added_group_dependencies: dict[str, list[str]] = {}
        self.uninstalled_dependencies: list[str] = []
        self.uninstalled_group_dependencies: dict[str, list[str]] = {}

    def manage_dependency(
        self, action: str, package: str, group: str | None = None
    ) -> None:
        data = self._parser.load()
        if not group:
            if action == "add":
                data["project"]["dependencies"] += [package]
                self.added_dependencies.append(package)
            else:
                data["project"]["dependencies"].remove(package)
                self.uninstalled_dependencies.append(package)
        else:
            if action == "add":
                data["project"]["optional-dependencies"][group] += [package]
                self.added_group_dependencies[group] = (
                    self.added_group_dependencies.get("group", []) + [package]
                )
            else:
                data["project"]["optional-dependencies"][group].remove(package)
                self.uninstalled_group_dependencies[group] = (
                    self.uninstalled_group_dependencies.get("group", [])
                    + [package]
                )
        self._parser.dump(data)

    def get_extra_groups(self) -> list[str]:
        data = self._parser.load()
        return list(data["project"].get("optional-dependencies", {}).keys())

    def is_installable(self) -> bool:
        return True

    @property
    def version(self) -> str:
        return "0.0.0"

    def change_version(self, new_version: str) -> str:
        return "0.0.0"

    def bump_version(self, rule: str) -> str:
        return "0.0.0"


class DummyInstaller(BaseInstaller):
    def __init__(self, toml: BaseToml) -> None:
        self.installed_packages: list[str] = []
        self._toml = toml

    def install(self, package: str, *, editable: bool = True) -> None:
        self.installed_packages.append(package)

    def uninstall(self, package: str) -> None:
        raise NotImplementedError

    def sync(self, requirements_files: list[str]) -> None:
        data = self._toml.load()
        requirements_per_file = {
            "requirements.lock": data["project"]["dependencies"],
            "requirements-dev.lock": data["project"]["optional-dependencies"][
                "dev"
            ],
            "requirements-test.lock": data["project"]["optional-dependencies"][
                "test"
            ],
        }
        packages: list[str] = []
        for f in requirements_files:
            packages.extend(requirements_per_file[f])
        self.installed_packages = packages


class DummyResolver(BaseResolver):
    def __init__(self) -> None:
        self.output_files: list[str] = []
        self.constraints_used: dict[str, str | None] = {}
        self.upgraded = False

    def compile(
        self,
        output_file: str,
        group: str | None = None,
        constraint_file: str | None = None,
        *,
        upgrade: bool = False,
    ) -> None:
        output_file = f"requirements{'-' + group if group else ''}.lock"
        self.output_files.append(output_file)
        self.constraints_used[output_file] = constraint_file
        self.upgraded = upgrade


class DummyVenv(BaseVirtualEnv):
    def __init__(self) -> None:
        self.created = False

    def already_created(self) -> bool:
        return self.created

    def create(self) -> None:
        self.created = True

    def get_path_to_command_bin(self, command: str) -> None:
        raise NotImplementedError


@pytest.fixture()
def requirements() -> dict[str, list[str]]:
    return {"main": ["foo", "bar"], "dev": ["developing"], "test": ["testing"]}


@pytest.fixture
def data(requirements: dict[str, list[str]]) -> dict[str, Any]:
    return {
        "project": {
            "name": "test",
            "dependencies": requirements["main"],
            "optional-dependencies": {
                "dev": requirements["dev"],
                "test": requirements["test"],
            },
        },
    }


@pytest.fixture
def toml(data: dict[str, Any]) -> BaseToml:
    return DummyToml(data)


@pytest.fixture
def pyproject(toml: BaseToml) -> BasePyproject:
    return DummyPyproject(toml)


@pytest.fixture
def installer(toml: BaseToml) -> BaseInstaller:
    return DummyInstaller(toml)


@pytest.fixture
def virtual_env() -> BaseVirtualEnv:
    return DummyVenv()


@pytest.fixture
def resolver() -> BaseResolver:
    return DummyResolver()


@pytest.fixture
def package_manager(
    pyproject: BasePyproject,
    installer: BaseInstaller,
    resolver: BaseResolver,
    virtual_env: BaseVirtualEnv,
) -> PackageManager:
    return PackageManager(pyproject, installer, resolver, virtual_env)


@pytest.fixture
def package() -> str:
    return "test-package"


def test_add_dependency(
    package: str,
    package_manager: PackageManager,
    pyproject: DummyPyproject,
    installer: DummyInstaller,
    resolver: DummyResolver,
) -> None:
    package_manager.manage_dependency("add", package)
    assert package in pyproject.added_dependencies
    assert package in installer.installed_packages
    assert "requirements.lock" in resolver.output_files


def test_remove_dependency(
    package_manager: PackageManager,
    pyproject: DummyPyproject,
    installer: DummyInstaller,
    resolver: DummyResolver,
) -> None:
    package = "foo"
    package_manager.manage_dependency("remove", package)
    assert package in pyproject.uninstalled_dependencies
    assert package not in installer.installed_packages
    assert "requirements.lock" in resolver.output_files


@pytest.mark.parametrize("action", ["add", "remove"])
def test_manage_dependency_compiles_all_files(
    action: Literal["add", "remove"],
    package_manager: PackageManager,
    resolver: DummyResolver,
) -> None:
    package = "bla" if action == "add" else "foo"
    package_manager.manage_dependency(action, package)

    groups = ["dev", "test"]
    files = ["requirements.lock"] + [
        f"requirements-{group}" for group in groups
    ]

    assert resolver.output_files.sort() == files.sort()


def test_add_dependency_with_group(
    package: str,
    package_manager: PackageManager,
    pyproject: DummyPyproject,
    installer: DummyInstaller,
    resolver: DummyResolver,
) -> None:
    group = "test"
    package_manager.manage_dependency("add", package, group)
    output_file = f"requirements-{group}.lock"

    assert package in pyproject.added_group_dependencies.get(group, [])
    assert package in installer.installed_packages
    assert output_file in resolver.output_files
    assert resolver.constraints_used[output_file] == "requirements.lock"


def test_remove_dependency_with_group(
    package_manager: PackageManager,
    pyproject: DummyPyproject,
    installer: DummyInstaller,
    resolver: DummyResolver,
) -> None:
    package = "testing"
    group = "test"
    package_manager.manage_dependency("remove", package, group)

    assert package in pyproject.uninstalled_group_dependencies.get(group, [])
    assert package not in installer.installed_packages
    assert f"requirements-{group}.lock" in resolver.output_files


def test_compile_requirements_upgrade(
    package_manager: PackageManager, resolver: DummyResolver
) -> None:
    package_manager.compile_requirements(upgrade=True)
    assert resolver.upgraded == True



from __future__ import annotations

from typing import Any

import pytest
from pspm.entities.toml import BaseToml
from pspm.entities.pyproject import Pyproject


class DummyToml(BaseToml):
    def __init__(self, data: dict[str, Any]) -> None:
        self.data = data

    def load(self) -> dict[str, Any]:
        return self.data

    def dump(self, data: dict[str, Any]) -> None:
        self.data = data


@pytest.fixture()
def requirements() -> dict[str, list[str]]:
    return {"main": ["foo", "bar"], "dev": ["developing"], "test": ["testing"]}


@pytest.fixture()
def version() -> str:
    return "1.1.1"


@pytest.fixture
def data(version: str, requirements: dict[str, list[str]]) -> dict[str, Any]:
    return {
        "project": {
            "name": "test",
            "version": version,
            "dependencies": requirements["main"],
            "optional-dependencies": {
                "dev": requirements["dev"],
                "test": requirements["test"],
            },
        },
    }


@pytest.fixture
def toml_parser(data: dict[str, Any]) -> BaseToml:
    return DummyToml(data)


@pytest.fixture()
def pyproject(toml_parser: BaseToml) -> Pyproject:
    return Pyproject(toml_parser)


def test_add_dependency(pyproject: Pyproject, toml_parser: BaseToml) -> None:
    package = "bla"
    data = toml_parser.load()
    expected_dependencies = data["project"]["dependencies"] + [package]
    pyproject.manage_dependency("add", package)
    result = toml_parser.load()
    assert result["project"]["dependencies"] == expected_dependencies


def test_remove_dependency(
    pyproject: Pyproject, toml_parser: BaseToml
) -> None:
    package = "foo"
    pyproject.manage_dependency("remove", package)
    result = toml_parser.load()
    assert result["project"]["dependencies"] == ["bar"]


def test_add_dependency_with_group(
    pyproject: Pyproject,
    toml_parser: BaseToml,
) -> None:
    package = "bla"
    group = "dev"
    data = toml_parser.load()
    expected_dependencies = data["project"].get(
        "optional-dependencies",
        {},
    ).get(group, []) + [
        package,
    ]
    pyproject.manage_dependency("add", package, group)
    result = toml_parser.load()
    assert (
        result["project"].get("optional-dependencies").get(group, [])
        == expected_dependencies
    )


def test_remove_dependency_with_group(
    pyproject: Pyproject, toml_parser: BaseToml
) -> None:
    group = "dev"
    pyproject.manage_dependency("remove", "developing", group)
    result = toml_parser.load()
    assert result["project"]["optional-dependencies"][group] == []


def test_dont_add_duplicate_dependency(
    pyproject: Pyproject, toml_parser: BaseToml
) -> None:
    data = toml_parser.load()
    dependencies = data["project"]["dependencies"].copy()
    package = dependencies[1]
    pyproject.manage_dependency("add", package)
    result = toml_parser.load()
    assert result["project"]["dependencies"] == dependencies


def test_get_extra_groups(pyproject: Pyproject, toml_parser: BaseToml) -> None:
    data = toml_parser.load()
    result = pyproject.get_extra_groups()
    assert result == ["dev", "test"]


def test_version_property(pyproject: Pyproject, version: str) -> None:
    assert pyproject.version == version


def test_version_change(pyproject: Pyproject) -> None:
    new_version = "2.3.4"
    pyproject.change_version(new_version)
    assert pyproject.version == new_version


@pytest.mark.parametrize(
    "rule,expected",
    [("major", "2.0.0"), ("minor", "1.2.0"), ("patch", "1.1.2")],
)
def test_version_bump(pyproject: Pyproject, rule: str, expected: str) -> None:
    pyproject.bump_version(rule)  # type: ignore
    assert pyproject.version == expected



"""Module for defining errors related to command execution."""


class CommandError(Exception):
    """Base error for command."""

    def __init__(self, command: str, message: str = "") -> None:
        """Initialize CommandError.

        Args:
            command: Command that failed to be executed.
            message: Error message
        """
        self.command = command
        self.message = message or f"Command {command} failed"
        super().__init__(self.message)


class CommandNotFoundError(CommandError):
    """Comand bin was not found."""


class CommandRunError(CommandError):
    """Command failed to run."""



"""Module to interact with command runner."""

from dotenv import load_dotenv
from rich import print as rprint

from pspm.entities.runner import Runner
from pspm.entities.virtual_env import VirtualEnv
from pspm.errors.command import CommandRunError


def _get_runner() -> Runner:
    virtual_env = VirtualEnv()
    return Runner(virtual_env)


def run_command(command: str, arguments: list[str]) -> None:
    """Load dotenv and run a command.

    Args:
        command: Command to run
        arguments: Arguments to pass to command
    """
    load_dotenv()
    runner = _get_runner()
    try:
        runner.run(command, arguments)
    except CommandRunError as e:
        rprint(f"[red]{e}[/red]")



"""Module for interacting with toml files."""

from __future__ import annotations

import abc
from pathlib import Path
from typing import Any

import tomli
import tomli_w


class BaseToml(abc.ABC):
    """TOML Parser and writer.

    Attributes:
        path: TOML file path
    """

    def __init__(self, path: str) -> None:
        """Initialize TOML.

        Args:
            path: File path to TOML file
        """
        self.path = path

    @abc.abstractmethod
    def load(self) -> dict[str, Any]:
        """Load TOML file."""
        raise NotImplementedError

    @abc.abstractmethod
    def dump(self, data: dict[str, Any]) -> None:
        """Write a dictionary to a file containing TOML-formatted data.

        Args:
            data: TOML data
        """
        raise NotImplementedError


class Toml(BaseToml):
    """TOML Parser and writer."""

    def load(self) -> dict[str, Any]:
        """Load TOML file.

        Returns:
            A dictionary containing parsed TOML
        """
        with Path(self.path).open("rb") as f:
            return tomli.load(f)

    def dump(self, data: dict[str, Any]) -> None:
        """Write a dictionary to a file containing TOML-formatted data.

        Args:
            data: TOML data
        """
        with Path(self.path).open("wb") as f:
            tomli_w.dump(data, f)



"""Module with classes to deal with packages."""

from __future__ import annotations

import abc
import subprocess

from pspm.errors.dependencies import InstallError, SyncError
from pspm.utils.bin_path import get_uv_path


class BaseInstaller(abc.ABC):
    """Package installer."""

    @abc.abstractmethod
    def install(self, package: str, *, editable: bool = False) -> None:
        """Install package.

        Args:
            package: Package to install
            editable: Whether to install in editable mode
        """
        raise NotImplementedError

    @abc.abstractmethod
    def uninstall(self, package: str) -> None:
        """Uninstall package.

        Args:
            package: Package to uninstall
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sync(self, requirements_files: list[str]) -> None:
        """Sync an environment with requirements files.

        Install all dependencies listed in requirements files
        and removes the ones that are not listed.

        Args:
            requirements_files: Path to files containing requirements
        """
        raise NotImplementedError


class UVInstaller(BaseInstaller):
    """Install packages with UV."""

    def __init__(self) -> None:
        """Initialize UV Installer."""
        self._uv_path = get_uv_path()

    def install(self, package: str, *, editable: bool = False) -> None:
        """Install a package.

        Args:
            package: Package to install
            editable: Whether to install in editable mode

        Raises:
            InstallError: If can't install package.
        """
        retcode = subprocess.call([
            self._uv_path,
            "pip",
            "install",
            *(["--editable"] if editable else []),
            package,
        ])
        if retcode != 0:
            raise InstallError(package)

    def uninstall(self, package: str) -> None:
        """Uninstall package.

        Args:
            package: Package to uninstall
        """
        raise NotImplementedError

    def sync(self, requirements_files: list[str]) -> None:
        """Sync an environment with requirements files.

        Install all dependencies listed in requirements files
        and removes the ones that are not listed.

        Args:
            requirements_files: Path to files containing requirements

        Raises:
            SyncError: If cant sync dependencies
        """
        retcode = subprocess.call([
            self._uv_path,
            "pip",
            "sync",
            *requirements_files,
        ])
        if retcode != 0:
            raise SyncError



"""Module to define utils functions that execute commands."""

from __future__ import annotations

import subprocess
from shutil import which


def get_git_user() -> dict[str, str] | None:
    """Retrieve git user.

    Returns:
        Git user with name and email
    """
    user: dict[str, str] = {}
    git_path = which("git")
    if not git_path:
        return None
    for field in ["name", "email"]:
        try:
            output = subprocess.check_output([
                git_path,
                "config",
                f"user.{field}",
            ])
        except subprocess.CalledProcessError:
            return None
        user[field] = output.strip().decode()
    return user



"""Module to define virtualenv classes."""

import abc
import subprocess
from pathlib import Path

from pspm.errors.command import CommandNotFoundError
from pspm.utils.bin_path import get_uv_path


class BaseVirtualEnv(abc.ABC):
    """Interacts with virtualenv."""

    @abc.abstractmethod
    def already_created(self) -> bool:
        """Check if env is already created."""
        raise NotImplementedError

    @abc.abstractmethod
    def create(self) -> None:
        """Create virtualenv."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_path_to_command_bin(self, command: str) -> str:
        """Retrieve path to command bin.

        Args:
            command: Command to find bin
        """
        raise NotImplementedError


class VirtualEnv(BaseVirtualEnv):
    """Interacts with virtualenv."""

    def __init__(self) -> None:
        """Initialize BaseVirtualEnv."""
        self._path = ".venv"

    def already_created(self) -> bool:
        """Check if env is already created.

        Returns:
            Whether the env is created
        """
        p = Path(self._path)
        return p.exists()

    def create(self) -> None:
        """Create virtualenv."""
        uv_path = get_uv_path()
        subprocess.run([uv_path, "venv", self._path], check=False)

    def get_path_to_command_bin(self, command: str) -> str:
        """Retrieve path to command bin.

        Args:
            command: Command to find bin path

        Raises:
            CommandNotFoundError: If command bin was not found

        Returns:
            Absolute path to command bin
        """
        command_path = Path(self._path) / "bin" / command
        if not command_path.exists():
            error_message = (
                f"Command {command} was not found "
                f"in {command_path.parent.absolute()}"
            )
            raise CommandNotFoundError(command, error_message)
        return str(command_path.absolute())



"""Module to define command runner class."""

import subprocess

from pspm.entities.virtual_env import VirtualEnv
from pspm.errors.command import CommandNotFoundError, CommandRunError


class Runner:
    """Class to run commands."""

    def __init__(self, virtual_env: VirtualEnv) -> None:
        """Initialize Runner.

        Args:
            virtual_env: Virtual env
        """
        self._virtual_env = virtual_env

    def run(self, command: str, arguments: list[str]) -> None:
        """Run a command.

        Args:
            command: Command to execute
            arguments: Arguments to be passed to command.

        Raises:
            CommandRunError: If command fails
        """
        try:
            self._virtual_env.get_path_to_command_bin(command)
        except CommandNotFoundError as e:
            raise CommandRunError(command, str(e)) from e

        retcode = subprocess.call([command, *arguments], shell=False)
        if retcode != 0:
            raise CommandRunError(command)



"""Module to define cli commands."""

# ruff: noqa: FBT001 FBT002 B006 UP007
from __future__ import annotations

import importlib.metadata
from enum import Enum
from pathlib import Path
from typing import Annotated, Optional

import typer
from rich import print as rprint

from pspm.services.bootstrap import bootstrap_project
from pspm.services.dependencies import (
    change_version,
    get_version,
    install_dependencies,
    lock_dependencies,
    manage_dependency,
)
from pspm.services.run import run_command

app = typer.Typer(no_args_is_help=True, invoke_without_command=True)


def _version_callback(value: bool) -> None:
    if value:
        version = importlib.metadata.version("pspm")
        rprint(version)


@app.callback()
def callback(
    version: Annotated[  # noqa: ARG001
        Optional[bool],
        typer.Option("--version", callback=_version_callback, is_eager=True),
    ] = None,
) -> None:
    """Python simple package manager."""


@app.command()
def init(
    path: Annotated[Path, typer.Argument(file_okay=False)] = Path(),
    name: Optional[str] = None,
    description: Optional[str] = None,
    installable: Annotated[
        bool, typer.Option("--installable/--not-installable")
    ] = True,
) -> None:
    """Create initial project structure.

    Args:
        path: Where to place the project
        name: Project name
        description: Project description
        installable: Whether the project is instalabble
    """
    bootstrap_project(path, name, description, installable=installable)


@app.command()
def install() -> None:
    """Install all dependencies and the package itself."""
    rprint("Installing project")
    install_dependencies()


@app.command()
def add(
    package: str,
    group: Annotated[Optional[str], typer.Option("--group", "-g")] = None,
) -> None:
    """Add package to pyproject, install it and lock version."""
    rprint(f"Adding package {package}")
    manage_dependency("add", package, group)


@app.command()
def remove(
    package: str,
    group: Annotated[Optional[str], typer.Option("--group", "-g")] = None,
) -> None:
    """Remove package from pyproject, uninstall it and lock version."""
    rprint(f"Removing package {package}")
    manage_dependency("remove", package, group or None)


@app.command()
def run(
    command: str,
    arguments: Annotated[Optional[list[str]], typer.Argument()] = None,
) -> None:
    """Run a command installed in virtual env.

    Args:
        command: Command to run
        arguments: Arguments to pass to command
    """
    run_command(command, arguments or [])


@app.command()
def lock(update: bool = False) -> None:
    """Lock the dependencies without installing."""
    lock_dependencies(update=update)


class BumpRules(str, Enum):  # noqa: D101
    major = "major"
    minor = "minor"
    patch = "patch"


@app.command()
def version(
    new_version: Annotated[Optional[str], typer.Argument()] = None,
    bump: Annotated[
        Optional[BumpRules],
        typer.Option("-b", "--bump"),
    ] = None,
) -> None:
    """Checks or update project version.

    Args:
        new_version: Version to change to
        bump: Bump rule to apply change version
    """
    if new_version:
        version = change_version(new_version)
    elif bump:
        version = change_version(bump_rule=bump.value)
    else:
        version = get_version()
    rprint(version)



"""Module with classes to deal with dependecy versions."""

from __future__ import annotations

import abc
import subprocess

from pspm.errors.dependencies import ResolveError
from pspm.utils.bin_path import get_uv_path


class BaseResolver(abc.ABC):
    """Base class for resolving dependencies.

    Attributes:
        pyproject_path: Path to pyproject.
    """

    def __init__(self, pyproject_path: str) -> None:
        """Initialize BaseResolver."""
        self.pyproject_path = pyproject_path

    @abc.abstractmethod
    def compile(
        self,
        output_file: str,
        group: str | None = None,
        constraint_file: str | None = None,
        *,
        upgrade: bool = False,
    ) -> None:
        """Compiles requirements into a lock file.

        Args:
            output_file: File to write output
            group: Group to include dependencies from
            constraint_file: Requirements file to contrain versions
            upgrade: Whether to upgrade package versions
        """
        raise NotImplementedError


class UVResolver(BaseResolver):
    """Class for resolving dependencies with UV."""

    def __init__(self, pyproject_path: str) -> None:
        """Initialize UV Compiler."""
        self.pyproject_path = pyproject_path
        self._uv_path = get_uv_path()

    def compile(
        self,
        output_file: str,
        group: str | None = None,
        constraint_file: str | None = None,
        *,
        upgrade: bool = False,
    ) -> None:
        """Compiles requirements into a lock file.

        Args:
            output_file: File to write output
            group: Group to include dependencies from
            constraint_file: Requirements file to contrain versions
            upgrade: Whether to upgrade package versions

        Raises:
            ResolveError: If cant resolve dependencies
        """
        retcode = subprocess.call([
            self._uv_path,
            "pip",
            "compile",
            "-q",
            *(["--extra", group] if group else []),
            *(["--constraint", constraint_file] if constraint_file else []),
            *(["--upgrade"] if upgrade else []),
            "-o",
            output_file,
            self.pyproject_path,
        ])
        if retcode != 0:
            raise ResolveError



"""Module to interact with pyproject.toml file."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Literal, cast

if TYPE_CHECKING:
    from pspm.entities.toml import BaseToml


class BasePyproject(abc.ABC):
    """Manipulate pyproject.

    Attributes:
        toml_parser: Parser to be used for parsing TOML
    """

    @abc.abstractmethod
    def manage_dependency(
        self,
        action: Literal["add", "remove"],
        package: str,
        group: str | None = None,
    ) -> None:
        """Add or removes dependency from project.

        Args:
            action: Action to take can be either add or remove
            package: Package to manage
            group: Group that package belongs
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_extra_groups(self) -> list[str]:
        """Retrieve list of extra groups."""
        raise NotImplementedError

    @abc.abstractmethod
    def is_installable(self) -> bool:
        """Determine if project is installable.

        Returns:
            Whether the project is installable
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def version(self) -> str:
        """Retrieve project version.

        Returns:
            Project version.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def change_version(self, new_version: str) -> str:
        """Change project version.

        Args:
            new_version: Version to change to

        Returns:
            Updated version
        """
        raise NotImplementedError

    @abc.abstractmethod
    def bump_version(self, rule: Literal["major", "minor", "patch"]) -> str:
        """Bump project version.

        Args:
            rule: Rule to apply when bumping

        Returns:
            Updated version
        """
        raise NotImplementedError


class Pyproject(BasePyproject):
    """Class for manipulating pyproject.toml file."""

    def __init__(self, toml_parser: BaseToml) -> None:
        """Iniatilize Pyproject.

        Args:
            toml_parser: Parser to be used for parsing TOML
        """
        self._parser = toml_parser
        self._data = toml_parser.load()

    def manage_dependency(
        self,
        action: Literal["add", "remove"],
        package: str,
        group: str | None = None,
    ) -> None:
        """Add or removes dependency from project.

        Args:
            action: Action to take can be either add or remove
            package: Package to manage
            group: Group that package belongs
        """
        data = self._parser.load()
        project: dict[str, Any] = data.get("project", {})
        optional_dependencies: dict[str, list[str]] = project.get(
            "optional-dependencies", {}
        )
        packages: list[str] = (
            project.get("dependencies", [])
            if not group
            else optional_dependencies.get(group, [])
        )
        if (action == "add" and package in packages) or (
            action == "remove" and package not in packages
        ):
            return
        (
            packages.append(package)
            if action == "add"
            else packages.remove(package)
        )
        if not group:
            project["dependencies"] = packages
        else:
            optional_dependencies[group] = packages
            project["optional-dependencies"] = optional_dependencies
        data["project"] = project
        self._parser.dump(data)

    def get_extra_groups(self) -> list[str]:
        """Retrieve list of extra groups.

        Returns:
            List of extra groups
        """
        data = self._parser.load()
        optional_dependencies = data["project"].get(
            "optional-dependencies",
            {},
        )
        return list(optional_dependencies.keys())

    def is_installable(self) -> bool:
        """Determine if project is installable.

        Returns:
            Whether the project is installable
        """
        data = self._parser.load()
        return data.get("build-system") is not None

    @property
    def version(self) -> str:
        """Retrieve project version.

        Returns:
            Project version.
        """
        return cast(str, self._data.get("project", {}).get("version", "0.0.0"))

    def change_version(self, new_version: str) -> str:
        """Change project version.

        Args:
            new_version: Version to change to

        Returns:
            Updated version
        """
        data = self._parser.load()
        project: dict[str, Any] = data.get("project", {})
        project["version"] = new_version
        data["project"] = project
        self._parser.dump(data)
        return new_version

    def bump_version(self, rule: Literal["major", "minor", "patch"]) -> str:
        """Bump project version.

        Args:
            rule: Rule to apply when bumping

        Returns:
            Updated version
        """
        rule_position = {"major": 0, "minor": 1, "patch": 2}[rule]
        parts = [int(p) for p in self.version.split(".", 3)]
        parts[rule_position] += 1
        for i in range(rule_position + 1, len(parts), 1):
            parts[i] = 0
        new_version = ".".join([str(p) for p in parts])
        return self.change_version(new_version)



"""Module to define PackageManager class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from pspm.errors.dependencies import ResolveError

if TYPE_CHECKING:
    from pspm.entities.installer import BaseInstaller
    from pspm.entities.pyproject import BasePyproject
    from pspm.entities.resolver import BaseResolver
    from pspm.entities.virtual_env import BaseVirtualEnv


class PackageManager:
    """Manage project dependencies."""

    def __init__(
        self,
        pyproject: BasePyproject,
        installer: BaseInstaller,
        resolver: BaseResolver,
        virtual_env: BaseVirtualEnv,
    ) -> None:
        """Initialize PackageManager.

        Args:
            pyproject: BasePyproject to manipulate pyproject file
            installer: BaseInstaller to install depencies
            resolver: BaseResolver to resolve dependencies
            virtual_env: BaseVirtualEnv to manage venv
        """
        self._pyproject = pyproject
        self._installer = installer
        self._resolver = resolver
        self._virtual_env = virtual_env

        self._main_requirements_file = "requirements.lock"
        self._group_requirements_file = "requirements-{}.lock"

    def _get_group_requirements_files(self) -> list[str]:
        groups = self._pyproject.get_extra_groups()
        return [self._group_requirements_file.format(g) for g in groups]

    def install(self) -> None:
        """Install all dependencies and the package itself."""
        if not self._virtual_env.already_created():
            self._virtual_env.create()

        self._installer.sync([
            self._main_requirements_file,
            *self._get_group_requirements_files(),
        ])
        if self._pyproject.is_installable():
            self._installer.install(".", editable=True)

    def manage_dependency(
        self,
        action: Literal["add", "remove"],
        package: str,
        group: str | None = None,
    ) -> None:
        """Add dependency to pyproject.

        Args:
            action: Action to take can be either add or remove
            package: Package to install
            group: Group to insert package
        """
        self._pyproject.manage_dependency(action, package, group)
        try:
            self.compile_requirements()
        except ResolveError:
            self._pyproject.manage_dependency("remove", package, group)
            return
        self.install()

    def compile_requirements(self, *, upgrade: bool = False) -> None:
        """Compile all requirements files.

        Args:
            upgrade: Whether to upgrade package versions
        """
        groups = self._pyproject.get_extra_groups()
        self._resolver.compile(self._main_requirements_file, upgrade=upgrade)
        for file, group in zip(self._get_group_requirements_files(), groups):
            self._resolver.compile(
                file,
                group,
                constraint_file=self._main_requirements_file,
                upgrade=upgrade,
            )



"""Module with errors related to dependency management."""


class DependencyError(Exception):
    """Base error for dependencies."""


class InstallError(DependencyError):
    """Can't install package."""

    def __init__(self, package: str) -> None:
        """Initialize InstallError.

        Args:
            package: Package that failed to be installed
        """
        self.package = package
        self.message = f"Error installing package {package}"
        super().__init__(self.message)


class SyncError(DependencyError):
    """Can't sync dependencies."""

    def __init__(self) -> None:
        """Initialize SyncError."""
        super().__init__("Could not sync dependencies")


class ResolveError(DependencyError):
    """Can't resolve dependencies."""

    def __init__(self) -> None:
        """Initialize SyncError."""
        super().__init__("Error resolving dependencies")
